/*
* Copyright 2016 Huy Cuong Nguyen
* Copyright 2016 ZXing authors
*/
// SPDX-License-Identifier: Apache-2.0

#pragma once

namespace ZXing {
namespace Pdf417 {

enum class Compaction {
	AUTO,
	TEXT,
	BYTE,
	NUMERIC
};

} // Pdf417
} // ZXing
