/*
* Copyright 2016 Huy Cuong Nguyen
* Copyright 2016 ZXing authors
*/
// SPDX-License-Identifier: Apache-2.0

#pragma once

namespace ZXing::DataMatrix {

enum class SymbolShape {
	NONE,
	SQUARE,
	RECTANGLE,
};

} // namespace ZXing::DataMatrix
